import argparse as argparse
from argparse import HelpFormatter

class ACmd(object):
    """ACmd is an abstraction that helps to
    define a consistant minimal interface for
    each command in this CLI.
    When a class is a subclass of this you have the assurance:
    The class is-a command of some sort
    It has info for its Name, Description, Example Usage and Help
    It has a run_command function to execute
    """
    REQUIRED_FIELDS = ['cmd_name', 'cmd_description', 'cmd_help', 'cmd_usage']


    def __init__(self, sub_parser):
        # Raise an exception if a subclass does not set values for all of the required fields
        if not all(getattr(self, field.upper()) is not None for field in self.REQUIRED_FIELDS):
            raise Exception("Command did not implement all the required fields")

        self.parser = sub_parser.add_parser(self.CMD_NAME,
                            help=self.CMD_HELP,
                            formatter_class=HelpFormatter,
                            parents=[])
        self.initialise(self)


    def initialise(self):
        raise NotImplementedError()

    def run_command(self, args):
        raise NotImplementedError()



from capsule.abstractions import ACmd

import pathlib
import sys
from capsule.lib.deployer import Deployer
from capsule.lib.config_handler import get_config
from capsule.lib.logging_handler import LOG
from terra_sdk.client.lcd import LCDClient
from terra_sdk.core import Coins
import requests
import asyncio
sys.path.append(pathlib.Path(__file__).parent.resolve())


class DeployCmd(ACmd):

    CMD_NAME = "deploy"
    CMD_HELP = "Deploy a wasm contract artifact to a specified Terra Chain"
    CMD_USAGE = """
    $ capsule deploy -p ./my_contract.wasm -c columbus-5
    $ capsule deploy --path ./artifacts/my_contract.wasm --chain tequila-0004"""
    CMD_DESCRIPTION = "Helper tool which enables you to programatically deploy a Wasm contract artifact to a chain as a code object and instantiate it"

    def initialise(self):
        # Define usage and description
        self.parser.usage = self.CMD_USAGE
        self.parser.description = self.CMD_DESCRIPTION

        # Add any positional or optional arguments here
        self.parser.add_argument("-p", "--package",
                                 type=str,
                                 help="(required) Name of new or path to existing package")
        
        # Add any positional or optional arguments here
        self.parser.add_argument("-i", "--initmsg",
                                 type=str,
                                 default={},
                                 help="(Optional) The initialization message for the contract you are trying to deploy. Must be a json-like str")

        self.parser.add_argument("-c", "--chain",
                                 type=str,
                                 default="",
                                 help="(Optional) A chain to deploy too. Defaults to localterra")


        
        
    def run_command(self, args):
        
        """Schema:
            Read Mnemonic from env as well as host to deploy on 
            any specified chain/account 

            Prepare defaults for the above

            Perform a store call for the wasm contract that was specified

            Verify the contract was stored with a new API call

            Instantiate the code object into a contract 

            Return success. 
        """
        LOG.info("Starting deployment")
        # Setup the Deployer with its lcd, fcd urls as well as the desired chain.
        # config = asyncio.run(get_config())
        chain_url="https://tequila-lcd.terra.dev"
        chain_fcd_url="https://tequila-fcd.terra.dev"
        
        deployer = Deployer(client=LCDClient(
            url=chain_url, 
            chain_id=args.chain or "tequila-0004",
            gas_prices=Coins(requests.get(f"{chain_fcd_url}/v1/txs/gas_prices").json())))
        
        # # Attempt to store the provided package as a code object, the response will be a code ID if successful
        stored_code_id = asyncio.run(deployer.store_contract(contract_name="test", contract_path=args.package))
        # Instantiate a contract using the stored code ID for our contract bundle
        # and an init msg which will be different depending on the contract.
        instantiation_result = asyncio.run(deployer.instantiate_contract(stored_code_id, init_msg=args.initmsg))



    

        




import os
from capsule.lib.config_handler import get_config

async def get_mnemonic(strict=False):
    """Attempt to gather a mnemonic from one of the available sources
    First, if a mnemonic is defined in the env, use that.
    Next, check the config file for the secret 
    If no mnemonic can be found, optionally raise an Exception
    

    Args:
        strict (bool, optional): When set to true, if no mnemonic is found an exception is raised. Defaults to False.

    Returns:
        str: The mnemonic found either in the env or in the config file
    """
    if os.getenv("CAPSULE_MNEMONIC", False):
        return os.environ["CAPSULE_MNEMONIC"]

    config = await get_config()
    if config.get("deploy_info", {}).get("mnemonic", False):
        return config.get("deploy_info", {}).get("mnemonic", False)
    
    if strict:
        raise Exception("No Mnemonic was found either in the specified config file or in the environment. Strict mode is set to true")
    return None
    



    



"""Load configuration from .toml file."""
import toml
import os
from capsule.lib.logging_handler import LOG

DEFAULT_CONFIG_FILE_ENV_VAR = "CAPSULE_CONFIG_FILE"
DEFAULT_CONFIG_FILE_NAME = "config.toml"

def get_config_file(filename=None):
    """Attempts to get the location of 
    the capsule config. 
    One of 4 things occurs here, either:
    - A config location is gathered from the environment
    - A specified filename is provided and a config file with this name is created
    - A specified filename is provided and if a config file with this name if found in the current dir then this is used
    - A default config file is created in the ~/.capsule directory

    Args:
        filename ([str], optional): A specified filename to create for config. Defaults to None.

    Returns:
        [str]: [The config files path.]
    """

    # First check if there is a specified location in the env
    env_config_file = os.environ.get(DEFAULT_CONFIG_FILE_ENV_VAR, None)

    # If a config was specified in the env, this takes priority
    if env_config_file: return env_config_file

    # If a filename was provided use this as the config file name
    if filename: return os.path.abspath(os.path.expandvars(os.path.expanduser(filename)))

    # Otherwise use a default one located in the .capsule directory at the home dir
    config_file = os.path.expanduser(os.path.join("~", ".capsule", DEFAULT_CONFIG_FILE_NAME))
    capsule_dir = os.path.dirname(config_file)
    # Check if the capsule directory has been created and if not, create it.
    if not os.path.exists(capsule_dir):
        os.makedirs(capsule_dir)
    return config_file

async def get_config(config_path=None):
    """Simple function which takes a config_file
    and attempts to parse it as a toml config
    returning the parsed result as a dict
    """
    filename = get_config_file(filename=config_path)
    # Read toml file
    config = toml.load(filename)

    LOG.debug(f"Found these networks available: {config['networks']}")
    LOG.debug(f"Found this deployment info: {config['deploy_info']['mnemonic']}")

    return config



from capsule.parser import get_main_parser, get_subcommmand_parser
from capsule.cmds import DeployCmd, AVAILABLE_COMMANDS
from capsule.lib.logging_handler import LOG

import sys 

def main():
    """This is the main entrypoint
    for the capsule tool.
    To comment briefly on its structure:
    - A main parser is established. This becomes the main router when a user enters the `capsule` command.
    - A subparser is added to the main parser. This becomes the keeper of each command
    - Each command is initialized with the subparser as a parent.
    - The main parse has all of its args parsed. Doing this after the above steps ensures we have all needed args for the help menu.
    """
    main_parser = get_main_parser()

    sub_parser = get_subcommmand_parser(main_parser)
    LOG.debug(AVAILABLE_COMMANDS)
    for cmd in AVAILABLE_COMMANDS:
        cmd.__init__(cmd, sub_parser=sub_parser)
    
    try:
        # Parse the arguments
        args = main_parser.parse_args()

        if args.cmd is None:
            main_parser.print_help()
            sys.exit()
    except Exception as e:
        main_parser.print_help()
        sys.exit()

    # Loop the commands again searching for the user-specified command
    for cmd in AVAILABLE_COMMANDS: 
        # If one of the available commands is the specified one       
        if cmd.CMD_NAME == args.cmd:
            # Run it with the provided args
            cmd.run_command(cmd, args)

    
        


"""Once Place for logging stuff"""
import logging


def setup_logger(logger_name="capsule", extra_kwargs={}):
    """Simple wrapper to setup a logger and return it.
    Used by the LOG constant which is used through the project

    Args:
        logger_name (str, optional): The name for the logger. Defaults to "capsule".
        extra_kwargs (dict, optional): Any extra options to use with the logger. Defaults to {}.
    """
    logger = logging.getLogger(logger_name, **extra_kwargs)
    logger.setLevel(logging.INFO)
    # Setup a StreamHandler to give output to the logs
    handler = logging.StreamHandler()
    # Establish a log format for messages
    handler.setFormatter(logging.Formatter('[capsule:%(module)s] %(message)s'))
    # Add handler to logger
    logger.addHandler(handler)
    return logger

LOG = setup_logger()



import base64
import json
import requests

from terra_sdk.core import Coins
from terra_sdk.client.lcd import LCDClient, Wallet
from terra_sdk.key.mnemonic import MnemonicKey
from terra_sdk.util.contract import read_file_as_b64, get_code_id
from terra_sdk.core.auth import StdFee
from terra_sdk.core.wasm import MsgStoreCode, MsgInstantiateContract, MsgExecuteContract

import pathlib
import sys
from capsule.lib.credential_handler import get_mnemonic
from capsule.lib.logging_handler import LOG
import asyncio
sys.path.append(pathlib.Path(__file__).parent.resolve())


class Deployer():
    """Deployer is a simple facade object
    providing an interface towards general deployment
    actions such as sending messages, getting addresses,
    uploading code objects, instantiating them into contracts
    and also executing or querying those contracts
    """

    def __init__(self, client: LCDClient) -> None:
        
        self.client = client
        self.mnemonic = asyncio.run(get_mnemonic())
        LOG.debug(self.mnemonic)
        self.deployer = Wallet(lcd=self.client, key=MnemonicKey(self.mnemonic))
        self.std_fee = StdFee(4000000, "600000uluna")

    async def send_msg(self, msg):
        """send_msg attempts to create 
        and sign a transaction with the provided
        msg and then broadcasts the tx

        """
        tx = self.deployer.create_and_sign_tx(
            msgs=[msg], fee=self.std_fee
        )
        # estimated = self.client.tx.estimate_fee(tx, fee_denoms=["uusd"], msgs=[msg])
        # LOG.info(f'estimated fee: {estimated}')
        return self.client.tx.broadcast(tx)

    def get_contract_address(instantiation_result):
        """get_contract_address takes a contract result
        and attempts to gather the contract address
        TODO: Refactor to make more general

        Args:
            instantiation_result ([type]): [description]

        Returns:
            str: The contracts address
        """
        log = json.loads(instantiation_result.raw_log)
        
        contract_address = ''
        for entry in log[0]['events'][0]['attributes']:
            if entry['key'] == 'contract_address':
                contract_address = entry['value']
        return contract_address

    async def store_contract(self, contract_name:str, contract_path:str="") -> str:
        """store_contract attempts to 
        gather a given wasm artifact file 
        and upload it to the given chain as a StoredCodeObject
        The storage operation is done using the send_msg helper

        Args:
            contract_name (str): The name of the contract to deploy
            contract_path (str, optional): The path to a wasm contract artifact to deploy. Defaults to "".

        Returns:
            str: The contract storage result is parsed and just the code id of the stored code object is returned
        """
        # If the full path was provided, use it else assume its located in artifacts
        bytes = read_file_as_b64(contract_path if contract_path else f"artifacts/{contract_name}.wasm")
        msg = MsgStoreCode(self.deployer.key.acc_address, bytes)
        contract_storage_result = await self.send_msg(msg)
        LOG.info(contract_storage_result)
        return get_code_id(contract_storage_result)
    
    async def instantiate_contract(self, code_id: str, init_msg) -> str:
        """instantiate_contract attempts to 
        instantiate a code object with an init msg 
        into a live contract on the network. 

        Args:
            code_id (str): The code_id of the stored wasm contract artifact
            init_msg (dict): The init msg to send to setup the contract

        Returns:
            str: The contracts address
        """
        msg = MsgInstantiateContract(
            owner=self.deployer.key.acc_address,
            code_id=code_id,
            init_msg=init_msg
        )
        instantiation_result = await self.send_msg(msg)
       
        LOG.debug(instantiation_result)
        return self.get_contract_address(instantiation_result)



from setuptools import setup, find_packages

from os import path
import io
this_directory = path.abspath(path.dirname(__file__))

with io.open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    readme_text = f.read()
    long_description = readme_text


setup(
    name="capsule_terra",
    license="MIT",
    version="0.0.2",
    packages=find_packages(),
    include_package_data=True,
    # Add command line: capsule and terra_deploy
    entry_points={
        "console_scripts": ["capsule=capsule.cli:main",
        "terra_deploy=capsule.cli:main"]
    },

    # PyPI metadata
    author="0xFable",
    author_email="0xfable@protonmail.com",
    description="Capsule is small Python SDK tool you can use to deploy Terra CosmWasm contracts to a given Terra (Cosmos SDK Based) chain.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        'argparse',
        'requests',
        'terra_sdk',
        'toml'
    ],
    keywords="cosmwasm terra",
)


from argparse import ArgumentParser, HelpFormatter


def get_main_parser():
    parser = ArgumentParser(
            prog="capsule",
            description="Python CLI which abstracts the deployment process into a small operation. Specify a config, run the deploy command while passing the location of the contract to deploy and let capsule handle the rest.",
            epilog="For support, please open a github issue.")

    parser.usage = """
    $ capsule <subcommand> ...
    $ capsule -v <subcommand> ...
    $ capsule -h
    """

    return parser

def get_subcommmand_parser(main_parser):
    # The subparser will act as a router for commands. 
    sub_parser = main_parser.add_subparsers(
        title="subcommands",
        description="one of these subcommands must be provided",
        metavar="",
        dest="cmd"
    )

    return sub_parser

